from itertools import cycle

import numpy as np
import pandas as pd

from agora.abc import ParametersABC, ProcessABC


class aggregateParameters(ParametersABC):
    """
    Parameters
    reduction: str to be passed to a dataframe for collapsing across columns
    """

    def __init__(self, reductions, axis, ranges):
        super().__init__()
        self.reductions = reductions
        self.axis = axis
        self.ranges = ranges

    @classmethod
    def default(cls):
        return cls.from_dict(
            {"reductions": ["mean", "median", "max"], "axis": 1, "ranges": None}
        )


class aggregate(ProcessABC):
    """
    Aggregate multiple datasets for cell-to-cell feature comparison.
    """

    def __init__(self, parameters: aggregateParameters):
        super().__init__(parameters)

    def run(self, signals):
        names = np.array([signal.index.names for signal in signals])
        index = signals[0].index
        for s in signals[0:]:
            index = index.intersection(s.index)

        tmp_signals = [s.loc[index] for s in signals]
        for i, s in enumerate(signals):
            tmp_signals[i].name = s.name
        signals = tmp_signals

        assert len(signals), "Signals is empty"

        bad_words = {
            "postprocessing",
            "extraction",
            "None",
            "np_max",
            "",
        }
        get_keywords = lambda df: [
            ind
            for item in df.name.split("/")
            for ind in item.split("/")
            if ind not in bad_words
        ]
        colnames = [
            "_".join(get_keywords(s) + [red])
            for s in signals
            for red in self.parameters.reductions
        ]
        concat = pd.concat(
            [
                getattr(signal, red)(axis=self.parameters.axis)
                for signal in signals
                for red in self.parameters.reductions
            ],
            names=signals[0].index.names,
            axis=self.parameters.axis,
        )
        if self.parameters.axis:
            concat.columns = colnames
        else:
            concat.columns = pd.MultiIndex.from_product(
                (
                    colnames,
                    [
                        "_".join((str(start), str(stop)))
                        for x in self.parameters.ranges
                        for start, stop in x
                    ],
                )
            )

        return concat



from typing import Dict, List, Union
import re

import numpy as np
import pandas as pd
from pandas import Series, DataFrame
from sklearn.cluster import KMeans
from matplotlib import pyplot as plt
import seaborn as sns

import compress_pickle

from postprocessor.core.postprocessor import PostProcessor
from postprocessor.core.tracks import get_avg_grs, non_uniform_savgol
from postprocessor.core.tracks import clean_tracks, merge_tracks, join_tracks
from postprocessor.core.ph import *

sns.set_context("talk", font_scale=1.8)
sns.set_theme(style="whitegrid")

# pp_c = PostProcessor(source=19920)  # 19916
# pp_c.load_tiler_cells()
# # f = "/home/alan/Documents/libs/extraction/extraction/examples/gluStarv_2_0_x2_dual_phl_ura8_00/extraction"
# # f = "/home/alan/Documents/tmp/pH_calibration_dual_phl__ura8__by4741__01/extraction/"
# f = "/home/alan/Documents/tmp/pH_calibration_dual_phl__ura8__by4741_Alan4_00/extraction/"
# pp_c.load_extraction(f)

# calib = process_phs(pp_c)
# # c = calib.loc[(5.4 < calib["ph"]) & (calib["ph"] < 8)].groupby("ph").mean()
# sns.violinplot(x="ph", y="em_ratio_mean", data=calib)
# plt.show()


# bring timelapse data and convert it to pH

pp = PostProcessor(source=19831)  # 19831
pp.load_tiler_cells()
f = "/home/alan/Documents/tmp/gluStarv_2_0_x2_dual_phl_ura8_00/extraction/"
# f = "/home/alan/Documents/tmp/downUpshift_2_0_2_glu_dual_phluorin__glt1_psa1_ura7__thrice_00/extraction/"
pp.load_extraction(f)

import compress_pickle

compress_pickle.dump(pp.extraction, "/home/alan/extraction_example.pkl")

if True:  # Load extracted data or pkld version
    new_dfs = compress_pickle.load("/home/alan/Documents/tmp/new_dfs.gz")
# Combine dataframes
else:
    new_dfs = combine_dfs(get_dfs(pp))
    # t = [x.index for x in new_dfs.values()]
    # i = set(t[0])
    # for j in t:
    #     i = i.intersection(j)
    new_dfs = {
        k: v
        for k, v in new_dfs.items()
        if k[2] != "imBackground"
        and k[2] != "median"
        and ~(((k[0] == "GFPFast") | (k[0] == "pHluorin405")) and k[2] == "max2p5pc")
    }

del pp
compress_pickle.dump(new_dfs, "/home/alan/Documents/tmp/new_dfs.gz")


def get_clean_dfs(dfs=None):
    if dfs is None:
        clean_dfs = compress_pickle.load("/home/alan/Documents/tmp/clean_dfs.gz")
    else:

        # Clean timelapse
        clean = clean_tracks(new_dfs[("general", None, "area")])
        tra, joint = merge_tracks(clean)
        clean_dfs = new_dfs
        i_ids = set(clean.index).intersection(
            clean_dfs[("general", None, "area")].index
        )
        clean_dfs = {k: v.loc[i_ids] for k, v in clean_dfs.items()}
        clean_dfs = {k: join_tracks(v, joint, drop=True) for k, v in clean_dfs.items()}

        del new_dfs
        compress_pickle.dump(clean_dfs, "/home/alan/Documents/tmp/clean_dfs.gz")


def plot_ph_hmap(clean_dfs):
    GFPFast = clean_dfs[("GFPFast", np.maximum, "mean")]
    phluorin = clean_dfs[("pHluorin405", np.maximum, "mean")]
    ph = GFPFast / phluorin
    ph = ph.loc[ph.notna().sum(axis=1) > 0.7 * ph.shape[1]]
    ph = 1 / ph

    fig, ax = plt.subplots()
    hmap(ph, cbar_kws={"label": r"emission ratio $\propto$ pH"})
    plt.xlabel("Time (hours)")
    plt.ylabel("Cells")
    xticks = plt.xticks(fontsize=15)[0]
    ax.set(yticklabels=[], xticklabels=[str(round(i * 5 / 60, 1)) for i in xticks])
    # plt.setp(ax.get_xticklabels(), Rotation=90)
    plt.show()


def fit_calibs(c, h):
    h = process_phs(pp_c)
    h["ratio"] = h["GFPFast_bgsub_median"] / h["pHluorin405_bgsub_median"]
    sns.lineplot(x="ph", y="ratio", data=h, err_style="bars")
    plt.show()

    calibs = fit_calibration(c)
    for k, params in calibs.items():
        i, j = ("_".join(k.split("_")[:-1]), k.split("_")[-1])
        if j == "mean" and "k2" not in k:
            clean_dfs[k] = objective(clean_dfs[i, np.maximum, j], *params)


# max 2.5% / med
def plot_ratio_vs_max2p5(h):
    fig, ax = plt.subplots()
    sns.regplot(
        x="em_ratio_median",
        y="mCherry_max2p5pc",
        data=h,
        scatter=False,
        ax=ax,
        color="teal",
    )
    sns.scatterplot(x="em_ratio_median", y="max5_d_med", data=h, hue="ph", ax=ax)
    plt.xlabel(r"Fluorescence ratio $R \propto (1/pH)$")
    plt.ylabel("Max 2.5% px / median")
    plt.show()


em = clean_dfs[("em_ratio", np.maximum, "mean")]
area = clean_dfs[("general", None, "area")]


def get_grs(clean_dfs):
    area = clean_dfs[("general", None, "area")]
    area = area.loc[area.notna().sum(axis=1) > 10]
    return area.apply(growth_rate, axis=1)


def get_agg(dfs, rng):
    # df dict of DataFrames containing an area/vol one TODO generalise this beyond area
    # rng tuple of section to use
    grs = get_grs(dfs)
    smooth = grs.loc(axis=1)[list(range(rng[0], rng[1]))].dropna(how="all")

    aggregate_mean = lambda dfs, rng: pd.concat(
        {
            k[0] + "_" + k[2]: dfs[k].loc[smooth.index, rng[0] : rng[1]].mean(axis=1)
            for k in clean_dfs.keys()
        },
        axis=1,
    )
    # f_comp_df = comp_df.loc[(comp_df["gr"] > 0) & (area.notna().sum(axis=1) > 50)]

    agg = aggregate_mean(dfs, rng)
    agg["max2_med"] = agg["mCherry_max2p5pc"] / agg["mCherry_mean"]

    for c in agg.columns:
        agg[c + "_log"] = np.log(agg[c])

    agg["gr_mean"] = smooth.loc[set(agg.index).intersection(smooth.index)].mean(axis=1)
    agg["gr_max"] = smooth.loc[set(agg.index).intersection(smooth.index)].max(axis=1)

    return agg


def plot_scatter_fit(x, y, data, hue=None, xlabel=None, ylabel=None, ylim=None):
    fig, ax = plt.subplots()
    sns.regplot(x=x, y=y, data=data, scatter=False, ax=ax)
    sns.scatterplot(x=x, y=y, data=data, ax=ax, alpha=0.1, hue=hue)
    # plt.show()
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if ylim is not None:
        plt.ylim(ylim)

    fig.savefig(
        "/home/alan/Documents/sync_docs/drafts/third_year_pres/figs/"
        + str(len(data))
        + "_"
        + x
        + "_vs_"
        + y
        + ".png",
        dpi=200,
    )


from extraction.core.argo import Argo, annot_from_dset


def additional_feats(aggs):
    aggs["gr_mean_norm"] = aggs["gr_mean"] * 12
    aggs["log_ratio_r"] = np.log(1 / aggs["em_ratio_mean"])
    return aggs


def compare_methods_ph_calculation(dfs):
    GFPFast = dfs[("GFPFast", np.maximum, "mean")]
    phluorin = dfs[("pHluorin405", np.maximum, "mean")]
    ph = GFPFast / phluorin

    sns.scatterplot(
        dfs["em_ratio", np.maximum, "mean"].values.flatten(),
        ph.values.flatten(),
        alpha=0.1,
    )
    plt.xlabel("ratio_median")
    plt.ylabel("median_ratio")
    plt.title("Comparison of ph calculation")
    plt.show()


# get the number of changes in a bool list
nchanges = lambda x: sum([i for i, j in zip(x[:-2], x[1:]) if operator.xor(i, j)])



import numpy as np
import pandas as pd

from postprocessor.core.processes.base import ParametersABC, ProcessABC


class savgolParameters(ParametersABC):
    """
    Parameters

        window : int (odd)
            Window length of datapoints. Must be odd and smaller than x
        polynom : int
            The order of polynom used. Must be smaller than the window size
    """

    def __init__(self, window, polynom):
        self.window = window
        self.polynom = polynom

    @classmethod
    def default(cls):
        return cls.from_dict({"window": 3, "polynom": 2})


class savgol(ProcessABC):
    """
    Apply Savitzky-Golay filter (works with NaNs, but it might return
    NaN regions).
    """

    def __init__(self, parameters: savgolParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        savgol_on_srs = lambda x: self.non_uniform_savgol(
            x.index, x.values, self.parameters.window, self.parameters.polynom
        )
        return signal.apply(savgol_on_srs, 1).apply(pd.Series)

    @staticmethod
    def non_uniform_savgol(x, y, window: int, polynom: int):
        """
        Applies a Savitzky-Golay filter to y with non-uniform spacing
        as defined in x

        This is based on https://dsp.stackexchange.com/questions/1676/savitzky-golay-smoothing-filter-for-not-equally-spaced-data
        The borders are interpolated like scipy.signal.savgol_filter would do

        source: https://dsp.stackexchange.com/a/64313

        Parameters
        ----------
        x : array_like
            List of floats representing the x values of the data
        y : array_like
            List of floats representing the y values. Must have same length
            as x
        window : int (odd)
            Window length of datapoints. Must be odd and smaller than x
        polynom : int
            The order of polynom used. Must be smaller than the window size

        Returns
        -------
        np.array of float
            The smoothed y values
        """
        if len(x) != len(y):
            raise ValueError('"x" and "y" must be of the same size')

        if len(x) < window:
            raise ValueError("The data size must be larger than the window size")

        if type(window) is not int:
            raise TypeError('"window" must be an integer')

        if window % 2 == 0:
            raise ValueError('The "window" must be an odd integer')

        if type(polynom) is not int:
            raise TypeError('"polynom" must be an integer')

        if polynom >= window:
            raise ValueError('"polynom" must be less than "window"')

        half_window = window // 2
        polynom += 1

        # Initialize variables
        A = np.empty((window, polynom))  # Matrix
        tA = np.empty((polynom, window))  # Transposed matrix
        t = np.empty(window)  # Local x variables
        y_smoothed = np.full(len(y), np.nan)

        # Start smoothing
        for i in range(half_window, len(x) - half_window, 1):
            # Center a window of x values on x[i]
            for j in range(0, window, 1):
                t[j] = x[i + j - half_window] - x[i]

            # Create the initial matrix A and its transposed form tA
            for j in range(0, window, 1):
                r = 1.0
                for k in range(0, polynom, 1):
                    A[j, k] = r
                    tA[k, j] = r
                    r *= t[j]

            # Multiply the two matrices
            tAA = np.matmul(tA, A)

            # Invert the product of the matrices
            tAA = np.linalg.inv(tAA)

            # Calculate the pseudoinverse of the design matrix
            coeffs = np.matmul(tAA, tA)

            # Calculate c0 which is also the y value for y[i]
            y_smoothed[i] = 0
            for j in range(0, window, 1):
                y_smoothed[i] += coeffs[0, j] * y[i + j - half_window]

            # If at the end or beginning, store all coefficients for the polynom
            if i == half_window:
                first_coeffs = np.zeros(polynom)
                for j in range(0, window, 1):
                    for k in range(polynom):
                        first_coeffs[k] += coeffs[k, j] * y[j]
            elif i == len(x) - half_window - 1:
                last_coeffs = np.zeros(polynom)
                for j in range(0, window, 1):
                    for k in range(polynom):
                        last_coeffs[k] += coeffs[k, j] * y[len(y) - window + j]

        # Interpolate the result at the left border
        for i in range(0, half_window, 1):
            y_smoothed[i] = 0
            x_i = 1
            for j in range(0, polynom, 1):
                y_smoothed[i] += first_coeffs[j] * x_i
                x_i *= x[i] - x[half_window]

        # Interpolate the result at the right border
        for i in range(len(x) - half_window, len(x), 1):
            y_smoothed[i] = 0
            x_i = 1
            for j in range(0, polynom, 1):
                y_smoothed[i] += last_coeffs[j] * x_i
                x_i *= x[i] - x[-half_window - 1]

        return y_smoothed



import pkg_resources
from pathlib import Path, PosixPath
from typing import Union, List, Dict
from datetime import datetime

from compress_pickle import load, dump

from core.experiment import Experiment
from core.cells import Cells
from core.segment import Tiler
from core.io.matlab import matObject
from core.experiment import Experiment


class PostProcessor:
    '''
    Base class to perform feature extraction.
    :param parameters: Parameters class with channels, reduction functions and
        extraction functions to use.
    :param source: Origin of experiment, if int it is assumed from Omero, if str
        or Path
    '''
    def __init__(self, parameters=None, source: Union[int, str, Path] = None):
        # self.params = parameters
        if source is not None:
            if type(source) is int:
                self.expt_id = source
                self.load_expt(source, omero=True)
            elif type(source) is str or PosixPath:
                self.load_expt(source, omero=False)

    @property
    def channels(self):
        if not hasattr(self, '_channels'):
            if type(self.params.tree) is dict:
                self._channels = tuple(self.params.tree.keys())

        return self._channels

    @property
    def current_position(self):
        assert hasattr(self, 'expt'), 'No experiment loaded.'

        if hasattr(self, 'tiler'):
            assert self.expt.current_position.name == self.tiler.current_position

        return self.expt.current_position

    def load_expt(self, source: Union[int, str], omero: bool = False) -> None:
        if omero:
            self.expt = Experiment.from_source(
                self.expt_id,  #Experiment ID on OMERO
                'upload',  #OMERO Username
                'gothamc1ty',  #OMERO Password
                'islay.bio.ed.ac.uk',  #OMERO host
                port=4064  #This is default
            )
        else:
            self.expt = Experiment.from_source(source)
            self.expt_id = self.expt.exptID

    def load_tiler_cells(self) -> None:
        self.tiler = Tiler(self.expt)
        self.cells = Cells()
        self.cells = self.cells.from_source(
            self.expt.current_position.annotation)

    def get_pos_mo_bud(self):
        annot = self.expt._get_position_annotation(
            self.expt.current_position.name)
        matob = matObject(annot)
        m = matob["timelapseTrapsOmero"].get("cellMothers", None)
        if m is not None:
            ids = np.nonzero(m.todense())
            d = {(self.expt.current_position.name, i, int(m[i, j])): []
                 for i, j in zip(*ids)}
            for i, j, k in zip(*ids, d.keys()):
                d[k].append((self.expt.current_position.name, i, j + 1))
        else:
            print("Pos {} has no mother matrix".format(
                self.expt.current_position.name))
            d = {}

        return d

    def get_exp_mo_bud(self):
        d = {}
        for pos in self.expt.positions:
            self.expt.current_position = pos
            d = {**d, **self.get_pos_mo_bud()}

        self.expt.current_position = self.expt.positions[0]

        return d

    def load_extraction(self, folder=None) -> None:
        if folder is None:
            folder = Path(self.expt.name + '/extraction')

        self.extraction = {}
        for pos in self.expt.positions:
            try:
                self.extraction[pos] = load(folder / Path(pos + '.gz'))

            except:
                print(pos, ' not found')



from itertools import product

import numpy as np
import pandas as pd
from itertools import product
import igraph as ig
import leidenalg

from agora.abc import ParametersABC, ProcessABC


class leidenParameters(ParametersABC):
    """
    Parameters
    """

    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    def default(cls):
        return cls.from_dict({})


class leiden(ProcessABC):
    """
    leiden algorithm applied to a dataframe with features.
    """

    def __init__(self, parameters: leidenParameters):
        super().__init__(parameters)

    def run(self, features: pd.DataFrame):
        # Generate euclidean distance matrix
        distances = np.linalg.norm(features.values - features.values[:, None], axis=2)
        ind = [
            "_".join([str(y) for y in x[1:]]) for x in features.index.to_flat_index()
        ]
        source, target = zip(*product(ind, ind))
        df = pd.DataFrame(
            {"source": source, "target": target, "distance": distances.flatten()}
        )
        df = df.loc[df["source"] != df["target"]]
        g = ig.Graph.DataFrame(df, directed=False)

        part = leidenalg.find_partition(g, leidenalg.ModularityVertexPartition)



from scipy.signal import argrelmax, argrelmin
from postprocessor.core.processes.base import ParametersABC, ProcessABC


class PeaksParameters(ParametersABC):
    """
    Parameters
        type : str {minima,  maxima, "all"}. Determines which type of peaks to identify
        order : int Parameter to pass to scipy.signal.argrelextrema indicating
            how many points to use for comparison.
    """

    def __init__(self, type):
        self.type = type

    @classmethod
    def default(cls):
        return cls.from_dict({"type": "minima", "order": 3})


class Peaks(ProcessABC):
    """
    Identifies a signal sharply dropping.
    """

    def __init__(self, parameters: PeaksParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        """
        Returns a boolean dataframe with the same shape as the
        original signal but with peaks as true values.
        """
        peaks_mat = np.zeros_like(signal, dtype=bool)

        comparator = np.less if self.parameters.type is "minima" else np.greater
        peaks_ids = argrelextrema(new_df, comparator=comparator, order=order)
        peaks_mat[peak_ids] = True

        return pd.DataFrame(
            peaks_mat,
            index=signal.index,
            columns=signal.columns,
        )



import numpy as np
import pandas as pd
from collections import namedtuple
from scipy.signal import periodogram

from agora.abc import ParametersABC, ProcessABC


class fftParameters(ParametersABC):
    """
    Parameters for the 'fft' process.

    Attributes
    ----------
    sampling_period: float
        Sampling period of measurement values, in unit time.

    oversampling_factor: float
        Oversampling factor, which indicates how many times a signal is
        sampled over the Nyquist rate.  For example, if the oversampling
        factor is 2, the signal is sampled at 2 times the Nyquist rate.
    """

    def __init__(
        self,
        sampling_period,
        oversampling_factor,
    ):
        # super().__init__()
        self.sampling_period = sampling_period
        self.oversampling_factor = oversampling_factor

    @classmethod
    def default(cls):
        # Sampling period should be listed in the HDF5 metadata (i.e. the
        # 'time_settings/timeinterval' attribute, unit seconds).  When using
        # this processor in a typical routine, use the sampling period from
        # there rather than relying on this default value of 5 minutes.
        return cls.from_dict(
            {
                "sampling_period": 5,
                "oversampling_factor": 1,
            }
        )


class fft(ProcessABC):
    """
    Process to estimate power spectral density (classical/Fourier).

    Methods
    -------
    classical_periodogram(timeseries: array_like, sampling_period: float,
        oversampling_factor: float)
        Estimates the power spectral density using a periodogram.
    run(signal: pd.DataFrame)
        Estimates the power spectral density of a dataframe of time series.
    """

    def __init__(self, parameters: fftParameters):
        super().__init__(parameters)

    def classical_periodogram(
        self,
        timeseries,
        sampling_period,
        oversampling_factor,
    ):
        """
        Estimates the power spectral density using a periodogram.

        This is based on a fast Fourier transform.

        The power spectrum is normalised so that the area under the power
        spectrum is constant across different time series, thus allowing users
        to easily compare spectra across time series.  See:
        * Scargle (1982). Astrophysical Journal 263 p. 835-853
        * Glynn et al (2006). Bioinformatics 22(3) 310-316

        Parameters
        ---------
        timeseries: array_like
            Time series of measurement values.

        sampling_period: float
            Sampling period of measurement values, in unit time.

        oversampling_factor: float
            Oversampling factor, which indicates how many times a signal is
            sampled over the Nyquist rate.  For example, if the oversampling
            factor is 2, the signal is sampled at 2 times the Nyquist rate.

        Returns
        -------
        freqs: ndarray
            Array of sample frequencies, unit time-1.

        power: ndarray
            Power spectral density or power spectrum of the time series,
            arbitrary units.
        """
        freqs, power = periodogram(
            timeseries,
            fs=1 / (oversampling_factor * sampling_period),
            nfft=len(timeseries) * oversampling_factor,
            detrend="constant",
            return_onesided=True,
            scaling="spectrum",
        )
        # Required to correct units; units will be expressed in min-1 (or any other
        # unit)
        freqs = oversampling_factor * freqs
        # Normalisation (Scargle, 1982; Glynn et al., 2006)
        power = power * (0.5 * len(timeseries))
        # Normalisation by variance to allow comparing different time series
        # (Scargle, 1982)
        power = power / np.var(timeseries, ddof=1)
        return freqs, power

    def run(self, signal: pd.DataFrame):
        """
        Estimates the power spectral density of a dataframe of time series.

        This uses the classical periodogram.

        All NaNs are dropped as the Fourier transform used does not afford
        missing time points or time points with uneven spacing in the time
        dimension.  For time series with missing values, the Lomb-Scargle
        periodogram is suggested (processes/lsp.py)

        Parameters
        ----------
        signal: pandas.DataFrame
            Time series, with rows indicating individiual time series (e.g. from
            each cell), and columns indicating time points.

        Returns
        -------
        freqs_df: pandas.DataFrame
            Sample frequencies from each time series, with labels preserved from
            'signal'.

        power_df: pandas.DataFrame
            Power spectrum from each time series, with labels preserved from
            'signal'.
        """
        FftAxes = namedtuple("FftAxes", ["freqs", "power"])
        # Each element in this list is a named tuple: (freqs, power)
        axes = [
            FftAxes(
                *self.classical_periodogram(
                    timeseries=signal.iloc[row_index, :].dropna().values,
                    sampling_period=self.sampling_period,
                    oversampling_factor=self.oversampling_factor,
                )
            )
            for row_index in range(len(signal))
        ]

        freqs_df = pd.DataFrame([element.freqs for element in axes], index=signal.index)

        power_df = pd.DataFrame([element.power for element in axes], index=signal.index)

        return freqs_df, power_df



"""
Main data processing functions. The inputs for this functions must be data series.
"""

import numpy as np
from pandas import Series, DataFrames

def bg_sub(data:Series, background:float):
    return data - background

def ratio(data1:Series, data2:Series):
    return data1 / data2

def norm_tps(data:Series, tps:list, fun=np.nanmean):
    return data / fun(data.loc(axis=1)[tps])

def growth_rate(data:Series, alg=None, filt = 'savgol'):
    if alg is None:
        alg='standard'

    

    



import h5py
from typing import List, Dict, Union
from pydoc import locate

import numpy as np
import pandas as pd

from tqdm import tqdm

from agora.abc import ParametersABC
from agora.io.writer import Writer
from agora.io.signal import Signal
from agora.io.cells import Cells

from postprocessor.core.processes.merger import mergerParameters, merger
from postprocessor.core.processes.picker import pickerParameters, picker
from postprocessor.core.processes.lineageprocess import LineageProcessParameters


class PostProcessorParameters(ParametersABC):
    """
    Anthology of parameters used for postprocessing
    :merger:
    :picker: parameters for picker
    :processes: list processes:[objectives], 'processes' are defined in ./processes/
        while objectives are relative or absolute paths to datasets. If relative paths the
        post-processed addresses are used. The order of processes matters.

    """

    def __init__(
        self,
        targets={},
        parameters={},
        outpaths={},
    ):
        self.targets: Dict = targets
        self.parameters: Dict = parameters
        self.outpaths: Dict = outpaths

    def __getitem__(self, item):
        return getattr(self, item)

    @staticmethod
    def find_in_1st(string, lol):
        pass

    @classmethod
    def default(cls, kind=[]):
        targets = {
            "prepost": {
                "merger": "/extraction/general/None/area",
                "picker": ["/extraction/general/None/area"],
            },
            "processes": [
                [
                    "bud_metric",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "births",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "dsignal",
                    [
                        "/extraction/general/None/volume",
                        "/postprocessing/bud_metric/extraction_general_None_volume",
                    ],
                ],
                [
                    "aggregate",
                    [
                        [
                            "/extraction/general/None/volume",
                            "postprocessing/bud_metric/extraction_general_None_volume",
                            "postprocessing/dsignal/extraction_general_None_volume",
                            "postprocessing/dsignal/postprocessing_bud_metric_extraction_general_None_volume",
                        ],
                    ],
                ],
            ],
        }
        parameters = {
            "prepost": {
                "merger": mergerParameters.default(),
                "picker": pickerParameters.default(),
            }
        }
        outpaths = {}
        outpaths["aggregate"] = "/postprocessing/experiment_wide/aggregated/"

        if "ph_batman" in kind:
            targets["processes"]["bud_metric"].append(
                [
                    [
                        "/extraction/em_ratio/np_max/mean",
                        "/extraction/em_ratio/np_max/median",
                    ],
                ]
            )
            targets["processes"]["dsignal"].append(
                [
                    "/extraction/em_ratio/np_max/mean",
                    "/extraction/em_ratio/np_max/median",
                    "/extraction/em_ratio_bgsub/np_max/mean",
                    "/extraction/em_ratio_bgsub/np_max/median",
                    "/postprocessing/bud_metric/extraction_em_ratio_np_max_mean",
                    "/postprocessing/bud_metric/extraction_em_ratio_np_max_median",
                ]
            )
            targets["processes"]["aggregate"].append(
                [
                    [
                        "/extraction/em_ratio/np_max/mean",
                        "/extraction/em_ratio/np_max/median",
                        "/extraction/em_ratio_bgsub/np_max/mean",
                        "/extraction/em_ratio_bgsub/np_max/median",
                        "/extraction/gsum/np_max/median",
                        "/extraction/gsum/np_max/mean",
                        "postprocessing/bud_metric/extraction_em_ratio_np_max_mean",
                        "postprocessing/bud_metric/extraction_em_ratio_np_max_median",
                        "postprocessing/dsignal/postprocessing_bud_metric_extraction_em_ratio_np_max_median",
                        "postprocessing/dsignal/postprocessing_bud_metric_extraction_em_ratio_np_max_mean",
                    ]
                ],
            )

        return cls(targets=targets, parameters=parameters, outpaths=outpaths)


class PostProcessor:
    def __init__(self, filename, parameters):
        self.parameters = parameters
        self._filename = filename
        self._signal = Signal(filename)
        self._writer = Writer(filename)

        # self.outpaths = parameters["outpaths"]
        self.merger = merger(
            mergerParameters.from_dict(parameters["parameters"]["prepost"]["merger"])
        )

        self.picker = picker(
            pickerParameters.from_dict(parameters["parameters"]["prepost"]["picker"]),
            cells=Cells.from_source(filename),
        )
        self.classfun = {
            process: self.get_process(process)
            for process, _ in parameters["targets"]["processes"]
        }
        self.parameters_classfun = {
            process: self.get_parameters(process)
            for process, _ in parameters["targets"]["processes"]
        }
        self.targets = parameters["targets"]

    @staticmethod
    def get_process(process):
        """
        Dynamically import a process class from the 'processes' folder.
        Assumes process filename and class name are the same
        """
        return locate("postprocessor.core.processes." + process + "." + process)

    @staticmethod
    def get_parameters(process):
        """
        Dynamically import parameters from the 'processes' folder.
        Assumes parameter is the same name as the file with 'Parameters' added at the end.
        """
        return locate(
            "postprocessor.core.processes." + process + "." + process + "Parameters"
        )

    def run_prepost(self):
        """Important processes run before normal post-processing ones"""

        merge_events = self.merger.run(self._signal[self.targets["prepost"]["merger"]])

        with h5py.File(self._filename, "r") as f:
            prev_idchanges = self._signal.get_merges()

        changes_history = list(prev_idchanges) + [np.array(x) for x in merge_events]
        self._writer.write("modifiers/merges", data=changes_history)

        with h5py.File(self._filename, "a") as f:  # TODO Remove this once done tweaking
            if "modifiers/picks" in f:
                del f["modifiers/picks"]

        indices = self.picker.run(self._signal[self.targets["prepost"]["picker"][0]])

        mothers, daughters = np.array(self.picker.mothers), np.array(
            self.picker.daughters
        )

        multii = pd.MultiIndex.from_arrays(
            (
                np.append(mothers, daughters[:, 1].reshape(-1, 1), axis=1).T
                if daughters.any()
                else [[], [], []]
            ),
            names=["trap", "mother_label", "daughter_label"],
        )
        self._writer.write(
            "postprocessing/lineage",
            data=multii,
            overwrite="overwrite",
        )

        # apply merge to mother-daughter
        moset = set([tuple(x) for x in mothers])
        daset = set([tuple(x) for x in daughters])
        picked_set = set([tuple(x) for x in indices])
        with h5py.File(self._filename, "a") as f:
            merge_events = f["modifiers/merges"][()]
        multii = pd.MultiIndex(
            [[], [], []], [[], [], []], names=["trap", "mother_label", "daughter_label"]
        )
        self.lineage_merged = multii
        if merge_events.any():
            merged_moda = set([tuple(x) for x in merge_events[:, 0, :]]).intersection(
                set([*moset, *daset, *picked_set])
            )
            search = lambda a, b: np.where(
                np.in1d(
                    np.ravel_multi_index(a.T, a.max(0) + 1),
                    np.ravel_multi_index(b.T, a.max(0) + 1),
                )
            )

            for target, source in merge_events:
                if (
                    tuple(source) in moset
                ):  # update mother to lowest positive index among the two
                    mother_ids = search(mothers, source)
                    mothers[mother_ids] = (
                        target[0],
                        self.pick_mother(mothers[mother_ids][0][1], target[1]),
                    )
                if tuple(source) in daset:
                    daughters[search(daughters, source)] = target
                if tuple(source) in picked_set:
                    indices[search(indices, source)] = target

            self.lineage_merged = pd.MultiIndex.from_arrays(
                (np.append(mothers, daughters[:, 1].reshape(-1, 1), axis=1).T),
                names=["trap", "mother_label", "daughter_label"],
            )
        self._writer.write(
            "postprocessing/lineage_merged",
            data=self.lineage_merged,
            overwrite="overwrite",
        )

        self._writer.write(
            "modifiers/picks",
            data=pd.MultiIndex.from_arrays(
                indices.T if indices.any() else [[], []],
                names=["trap", "cell_label"],
            ),
            overwrite="overwrite",
        )

    @staticmethod
    def pick_mother(a, b):
        """Update the mother id following this priorities:

        The mother has a lower id
        """
        x = max(a, b)
        if min([a, b]):
            x = [a, b][np.argmin([a, b])]
        return x

    def run(self):
        self.run_prepost()

        for process, datasets in tqdm(self.targets["processes"]):
            if process in self.parameters["parameters"].get(
                "processes", {}
            ):  # If we assigned parameters
                parameters = self.parameters_classfun[process](self.parameters[process])

            else:
                parameters = self.parameters_classfun[process].default()

            if isinstance(parameters, LineageProcessParameters):
                with h5py.File(self._filename, "r") as f:
                    trap_mo_da = f[parameters.lineage_location]
                    lineage = np.array(
                        (
                            trap_mo_da["trap"],
                            trap_mo_da["mother_label"],
                            trap_mo_da["daughter_label"],
                        )
                    ).T
                loaded_process = self.classfun[process](parameters)
                loaded_process.load_lineage(lineage)
            else:
                loaded_process = self.classfun[process](parameters)

            for dataset in datasets:
                # print("Processing", process, "for", dataset)

                if isinstance(dataset, list):  # multisignal process
                    signal = [self._signal[d] for d in dataset]
                elif isinstance(dataset, str):
                    signal = self._signal[dataset]
                else:
                    raise ("Incorrect dataset")

                if len(signal):
                    result = loaded_process.run(signal)
                else:
                    result = pd.DataFrame(
                        [], columns=signal.columns, index=signal.index
                    )

                if process in self.parameters["outpaths"]:
                    outpath = self.parameters["outpaths"][process]
                elif isinstance(dataset, list):
                    # If no outpath defined, place the result in the minimum common
                    # branch of all signals used
                    prefix = "".join(
                        prefix + c[0]
                        for c in takewhile(
                            lambda x: all(x[0] == y for y in x), zip(*dataset)
                        )
                    )
                    outpath = (
                        prefix
                        + "_".join(  # TODO check that it always finishes in '/'
                            [d[len(prefix) :].replace("/", "_") for d in dataset]
                        )
                    )
                elif isinstance(dataset, str):
                    outpath = dataset[1:].replace("/", "_")
                else:
                    raise ("Outpath not defined", type(dataset))

                if process not in self.parameters["outpaths"]:
                    outpath = "/postprocessing/" + process + "/" + outpath

                if isinstance(result, dict):  # Multiple Signals as output
                    for k, v in result:
                        self.write_result(
                            outpath + f"/{k}",
                            v,
                            metadata={},
                        )
                else:
                    self.write_result(
                        outpath,
                        result,
                        metadata={},
                    )

    def write_result(
        self, path: str, result: Union[List, pd.DataFrame, np.ndarray], metadata: Dict
    ):
        self._writer.write(path, result, meta=metadata)


def _if_dict(item):
    if hasattr(item, "to_dict"):
        item = item.to_dict()
    return item



from pathlib import Path

from postprocessor.core.group import GroupParameters, Group

poses = [
    x.name.split("store")[0]
    for x in Path(
        "/shared_libs/pipeline-core/scripts/data/ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01"
    ).rglob("*")
    if x.name != "images.h5"
]

gr = Group(
    GroupParameters(
        signals=["/extraction/general/None/area", "/extraction/mCherry/np_max/median"]
    )
)
gr.run(
    central_store="/shared_libs/pipeline-core/scripts/data/ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01",
    poses=poses,
)



"""
Class to group multiple positions into one using one of several criteria.
"""

from pathlib import Path
import re

import h5py
import pandas as pd

from agora.io.bridge import groupsort
from postprocessor.core.io.signal import Signal

from postprocessor.core.processes.base import ParametersABC, ProcessABC


class GroupParameters(ParametersABC):
    def __init__(self, by="name", processes=[], signals=[]):
        self.by = by
        self.signals = signals
        self.processes = processes

    @classmethod
    def default(cls):
        return cls.from_dict({"by": "name", "signals": [], "processes": []})


class Group(ProcessABC):
    def __init__(self, parameters):
        super().__init__(parameters)

    def get_position_filenames(self, exp_root, poses):
        """
        Get filenames as a dictionary where the key is the position and value the filename.
        """
        central_store = Path(exp_root) / "store.h5"
        if central_store.exists():
            hdf = h5py.File(central_store, "r")
            self.filenames = [pos.attrs["filename"] for pos in hdf["/positions/"]]
            hdf.close()
        else:  # If no central store just list position files in expt root folder
            fullfiles = [x for x in Path(exp_root).glob("*store.h5")]
            files = [x.name for x in Path(exp_root).glob("*store.h5")]
            filenames = [False for _ in poses]
            for i, pos in enumerate(poses):
                matches = [
                    True if re.match(pos + ".*.h5", fname) else False for fname in files
                ]
                if any(matches):
                    assert sum(matches) == 1, "More than one match"
                    filenames[i] = (pos, fullfiles[matches.index(True)])

            self.filenames = {fname[0]: fname[1] for fname in filenames if fname}

        self.positions = list(self.filenames.keys())
        return self.filenames

    def get_signals(self):
        # hdf = h5py.File(central_store, "r")
        # keys_d = groupsort(keys)
        self.signals = {pos: {} for pos in self.filenames.keys()}
        for pos, fname in self.filenames.items():
            for signal in self.parameters.signals:
                self.signals[pos][signal] = pd.read_hdf(fname, signal)

        return self.signals

    def gen_groups(self):
        if self.by == "group":  # Use group names in metadata
            pass
        elif self.by == "name":  # Infer groups from signal concatenation
            # Remove last four characters and find commonalities larger than 4
            # characters between posnames and group them that way.
            groupnames = list(set([x[:-3] for x in self.positions]))
            self.group_signal_tree = {group: [] for group in groupnames}
            self.poses_grouped = {group: [] for group in groupnames}
            for pos in self.positions:
                group = groupnames[groupnames.index(pos[:-3])]
                self.group_signal_tree[group].append(self.signals[pos])
                self.poses_grouped[group].append(pos)

        elif (
            type(self.by) == tuple
        ):  # Manually give groups as tuple or list of positions
            pass

    def concat_signals(self):
        self.concated_signals = {group: {} for group in self.group_signal_tree}
        for k, group in self.group_signal_tree.items():
            for signal in self.parameters.signals:
                self.concated_signals[k][signal] = pd.concat(
                    [g[signal] for g in group], keys=self.poses_grouped[k]
                )

        return self.concated_signals

    def process_signals(self, grouped_signals):
        pass

    def run(self, central_store, poses):

        self.get_position_filenames(central_store, poses)
        self.get_signals()
        self.gen_groups()
        self.concat_signals()
        # processed_signals = self.process_signals(grouped_signals)

        return self.concated_signals
        # return processed_signals


poses = [
    x.name.split("store")[0]
    for x in Path(
        "/shared_libs/pipeline-core/scripts/data/ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01"
    ).rglob("*")
    if x.name != "images.h5"
]
gr = Group(
    GroupParameters(
        signals=["/extraction/general/None/area", "/extraction/mCherry/np_max/median"]
    )
)
gr.run(
    central_store="/shared_libs/pipeline-core/scripts/data/ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01",
    poses=poses,
)
signal = Signal(
    "/shared_libs/pipeline-core/scripts/data/ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01/ph_5_04_001store.h5"
)



#!/usr/bin/env python3


import numpy as np
import pandas as pd
from sklearn import decomposition
from catch22 import catch22_all

from agora.abc import ParametersABC, ProcessABC


class catch22Parameters(ParametersABC):
    """
    Parameters
        :min_len: Prefilter to account only for long-signal cells
    """

    def __init__(self, min_len, n_components):
        self.min_len = min_len
        self.n_components = n_components

    @classmethod
    def default(cls):
        return cls.from_dict(
            {
                "min_len": 0.8,
                "n_components": None,
            }
        )


class catch22(ProcessABC):
    """
    catch22 class. It produces 22 normalised features for each time lapse in the signal (using the catch22 library.)
    """

    def __init__(self, parameters: catch22Parameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        thresh = (
            self.min_len
            if isinstance(self.min_len, int)
            else signal.shape[1] * self.min_len
        )
        adf = signal.loc[signal.notna().sum(axis=1) > thresh]
        catches = [catch22_all(adf.iloc[i, :].dropna().values) for i in range(len(adf))]

        norm = pd.DataFrame(
            [catches[j]["values"] for j in range(len(catches))],
            index=adf.index,
            columns=catches[0]["names"],
        )

        return norm



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['postprocessor',
 'postprocessor.core',
 'postprocessor.core.functions',
 'postprocessor.core.old',
 'postprocessor.core.processes',
 'postprocessor.examples']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'aliby-agora>=0.2.11,<0.3.0',
 'catch22>=0.2.0,<0.3.0',
 'leidenalg>=0.8.8,<0.9.0',
 'more-itertools>=8.12.0,<9.0.0',
 'numpy>=1.17.3',
 'omni-gaussian',
 'pandas>=1.3.5,<2.0.0',
 'pathos>=0.2.8,<0.3.0',
 'scipy>=1.4.1']

setup_kwargs = {
    'name': 'aliby-post',
    'version': '0.1.9',
    'description': 'Post-processing tools for aliby pipeline.',
    'long_description': None,
    'author': 'Alán Muñoz',
    'author_email': 'amuoz@ed.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



from postprocessor.core.processor import PostProcessor, PostProcessorParameters

params = PostProcessorParameters.default()
pp = PostProcessor(
    "/shared_libs/pipeline-core/scripts/pH_calibration_dual_phl__ura8__by4741__01/ph_5_29_025store.h5",
    params,
)
tmp = pp.run()

import h5py

# f = h5py.File(
#     "/shared_libs/pipeline-core/scripts/pH_calibration_dual_phl__ura8__by4741__01/ph_5_29_025store.h5",
#     "a",
# )



#!/usr/bin/env python3


import pandas as pd

from agora.abc import ParametersABC, ProcessABC


class signalParameters(ParametersABC):
    """
    :window: Number of timepoints to consider for signal.
    """

    def __init__(self, window: int):
        self.window = window

    @classmethod
    def default(cls):
        return cls.from_dict({"window": 3})


class mednorm(ProcessABC):
    """
    Normalise a channel by the median
    """

    def __init__(self, parameters: mednormParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        return signal.rolling(window=self.parameters.window, axis=1).mean().diff(axis=1)



#!/usr/bin/env python3

"""
Main dataframe structure

| position | group | ntraps |robustness index | initial_ncells | final_ncells
"""
# dir = "/home/alan/Documents/dev/skeletons/data/2021_06_15_pypipeline_unit_test_00/2021_06_15_pypipeline_unit_test_00/"
# dir = "/home/alan/Documents/dev/libs/aliby/data/2021_08_24_2Raf_00/2021_08_24_2Raf_00/"
dirs = [
    # "2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01/",
    # "2020_10_01_exp_00/",
    # "2020_10_02_downUpshift_twice_2_0_2_glu_ura8_phluorinMsn2_phluorinMig1_01/",
    "/home/alan/Documents/dev/skeletons/data/2021_05_27_PDR5Fluc_00/2021_05_27_PDR5Fluc_00/",
    # "2020_10_22_downUpshift_2_0_2_glu_dual_phluorin__glt1_psa1_ura7__thrice_00/",
    # "2021_06_15_pypipeline_unit_test_00/",
    # "2021_06_27_ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01/",
    # "/home/alan/Documents/dev/skeletons/data/2021_08_24_2Raf_00/2021_08_24_2Raf_00",
    # "/home/alan/Documents/dev/skeletons/data/2021_09_15_1_Raf_06/2021_09_15_1_Raf_06/"
    # "/home/alan/Documents/dev/libs/aliby/data/sofia/2021_08_24_2Raf_00/2021_08_24_2Raf_00",
    # "/home/alan/Documents/dev/libs/aliby/data/sofia/2021_09_15_1_Raf_06/2021_09_15_1_Raf_06",
    # "/home/alan/Documents/dev/libs/aliby/data/sofia/2021_09_17_0_1_Raf_00/2021_09_17_0_1_Raf_00",
    "/home/alan/Documents/dev/skeletons/data/2021_09_15_1_Raf_06/2021_09_15_1_Raf_06/",
    "/home/alan/Documents/dev/skeletons/data/2021_08_24_2Raf_00/2021_08_24_2Raf_00/",
    "/home/alan/Documents/dev/skeletons/data/2021_11_01_01_Raf_00/2021_11_01_01_Raf_00/",
]
import h5py

from abc import abstractclassmethod, abstractmethod
from typing import Iterable, Union, Dict, Tuple
from pathlib import PosixPath, Path
import warnings
from collections import Counter

import numpy as np
from numpy import ndarray
import pandas as pd
from scipy.signal import find_peaks

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns

from agora.abc import ProcessABC, ParametersABC
from postprocessor.grouper import NameGrouper


# group_pos_trap_ncells = (
#     concat.dropna().groupby(["group", "position", "trap"]).apply(len)
# )
# group_pos_trapswcell = (
#     group_pos_trap_ncells.dropna().groupby(["group", "position"]).apply(len)
# )


class Meta:
    """
    Convenience class to fetch data from hdf5 file
    """

    def __init__(self, filename):
        self.filename = filename

    @property
    def ntimepoints(self):
        with h5py.File(self.filename, "r") as f:
            return f.attrs["time_settings/ntimepoints"][0]


class Compiler(ProcessABC):
    def __init__(self, parameters):
        pass
        # super().__init__(parameters)

    @abstractmethod
    def load_data(self):
        """Abstract function that must be reimplemented"""
        pass

    @abstractmethod
    def run():
        pass


class ExperimentCompiler(Compiler):
    def __init__(self, CompilerParameters, exp_path: PosixPath):
        super().__init__(CompilerParameters)
        self.load_data(exp_path)

    def run(self):

        return {
            method: getattr(self, "compile_" + method)()
            for method in ("slice", "slices", "delta_traps", "pertrap_metric")
        }

    def load_data(self, path: PosixPath):
        self.grouper = NameGrouper(path)
        self.meta = Meta(self.grouper.files[0])

    @property
    def ntraps(self) -> dict:
        """Get the number of traps in each position

        Returns
        -------
        dict str -> int

        Examples
        --------
        FIXME: Add docs.


        """
        return {pos: coords.shape[0] for pos, coords in self.grouper.traplocs().items()}

    def concat_signal(
        self, sigloc=None, mode="mothers", *args, **kwargs
    ) -> pd.DataFrame:

        if sigloc == None:
            sigloc = "extraction/general/None/volume"
        self.sigloc = sigloc

        if mode == None:
            mode = "mothers"

        if not hasattr(self, "_concat") or self.sigloc != sigloc:
            self._concat = self.grouper.concat_signal(
                self.sigloc, pool=7, mode="mothers"
            )

        return self._concat

    def get_tp(self, sigloc=None, tp=None, mode=None, *args, **kwargs) -> pd.Series:

        if tp is None:
            tp = 0

        if mode == None:
            mode = True

        return self.concat_signal(sigloc=sigloc, mode=mode, *args, **kwargs).iloc[:, tp]

    def compile_slices(self, nslices=2, *args, **kwargs):
        tps = [
            min(i * (self.meta.ntimepoints // nslices), self.meta.ntimepoints - 1)
            for i in range(nslices + 1)
        ]
        slices = [self.compile_slice(tp=tp) for tp in tps]
        slices_df = pd.concat(slices)

        slices_df["timepoint"] = np.concatenate(
            [np.repeat(tp, len(slice_df)) for tp, slice_df in zip(tps, slices)]
        )

        return slices_df

    def compile_slice_end(self, *args, **kwargs):
        return self.compile_slice(tp=-1, *args, **kwargs)

    def compile_slice(
        self, sigloc=None, tp=None, metrics=None, mode=None, *args, **kwargs
    ) -> pd.DataFrame:

        if sigloc == None:
            self.sigloc = "extraction/general/None/volume"

        if tp == None:
            tp = 0

        if metrics == None:
            metrics = ("max", "mean", "median", "count", "std", "sem")

        if mode == None:
            mode = True

        df = pd.concat(
            [
                getattr(
                    self.get_tp(sigloc=sigloc, tp=tp, mode=mode, *args, **kwargs)
                    .groupby(["group", "position", "trap"])
                    .max()
                    .groupby(["group", "position"]),
                    met,
                )()
                for met in metrics
            ],
            axis=1,
        )

        df.columns = metrics

        merged = self.add_column(df, self.ntraps, name="ntraps")

        return merged

    @staticmethod
    def add_column(df: pd.DataFrame, new_values_d: dict, name="new_col"):

        if name in df.columns:
            warnings.warn("ExpCompiler: Replacing existing column in compilation")
        df[name] = [new_values_d[pos] for pos in df.index.get_level_values("position")]

        return df.reset_index()

    @staticmethod
    def traploc_diffs(traplocs: ndarray) -> list:
        """Obtain metrics for trap localisation.

        Parameters
        ----------
        traplocs : ndarray
            (x,2) 2-dimensional array with the x,y coordinates of traps in each
            column

        Examples
        --------
        FIXME: Add docs.

        """
        signal = np.zeros((traplocs.max(), 2))
        for i in range(2):
            counts = Counter(traplocs[:, i])
            for j, v in counts.items():
                signal[j - 1, i] = v

        where_x = np.where(signal[:, 0])[0]
        where_y = np.where(signal[:, 1])[0]

        diffs = [
            np.diff(x)
            for x in np.apply_along_axis(find_peaks, 0, signal, distance=10)[0]
        ]
        return diffs

    def compile_delta_traps(self):
        group_names = compiler.grouper.group_names
        tups = [
            (group_names[pos], pos, axis, val)
            for pos, coords in self.grouper.traplocs().items()
            for axis, vals in zip(("x", "y"), self.traploc_diffs(coords))
            for val in vals
        ]

        return pd.DataFrame(tups, columns=["group", "position", "axis", "value"])

    def compile_pertrap_metrics(self):
        pass

    def compile_pertrap_metric(
        self,
        ranges: Iterable[Iterable[int]] = [
            [0, -1],
        ],
        metric: str = "count",
    ):
        "Get the number of cells per trap present during the given ranges"
        sig = compiler.concat_signal()

        for i, rngs in enumerate(ranges):
            for j, edge in enumerate(rngs):
                if edge < 0:
                    ranges[i][j] = sig.shape[1] - i + 1

        return pd.concat(
            [
                self.get_filled_trapcounts(sig.loc(axis=1)[slice(*rng)], metric=metric)
                for rng in ranges
            ],
            axis=1,
        )

    def get_filled_trapcounts(self, signal: pd.DataFrame, metric: str) -> pd.Series:
        present = signal.apply(
            lambda x: (not x.first_valid_index())
            & (x.last_valid_index() == len(x) - 1),
            axis=1,
        )
        results = getattr(
            signal.loc[present].iloc[:, 0].groupby(["group", "position", "trap"]),
            metric,
        )()
        filled = self.fill_trapcount(results)
        return filled

    def fill_trapcount(
        self, srs: pd.Series, fill_value: Union[int, float] = 0
    ) -> pd.Series:
        """Fill the last level of a MultiIndex in a pd.Series

        Use compiler to get the max number of traps per position and use this
        information to add rows with empty values (with plottings of distributions
        in mind)

        Parameters
        ----------
        srs : pd.Series
            Series with a pd.MultiIndex index
        compiler : ExperimentCompiler
            class with 'ntraps' information that returns a dictionary with position
            -> ntraps.
        fill_value : Union[int, float]
            Value used to fill new rows.

        Returns
        -------
        pd.Series
            Series with no numbers skipped on the last level.

        Examples
        --------
        FIXME: Add docs.

        """

        all_sets = set(
            [(pos, i) for pos, ntraps in compiler.ntraps.items() for i in range(ntraps)]
        )
        dif = all_sets.difference(
            set(
                zip(*[srs.index.get_level_values(i) for i in ("position", "trap")])
            ).difference()
        )
        new_indices = pd.MultiIndex.from_tuples(
            [
                (compiler.grouper.group_names[idx[0]], idx[0], np.uint(idx[1]))
                for idx in dif
            ]
        )
        new_indices = new_indices.set_levels(
            new_indices.levels[-1].astype(np.uint), level=-1
        )
        empty = pd.Series(fill_value, index=new_indices, name="ncells")
        return pd.concat((srs, empty))


# fig = plt.figure(tight_layout=True)
# gs = Grid_plot = plt.GridSpec(3, 2, wspace=0.8, hspace=0.6)

# sns.stripplot(
#     data=merged.reset_index(),
#     x="group",
#     y="count",
#     hue="ntraps",
#     ax=fig.add_subplot(gs[0, 0]),
# )

# g = NameGrouper(dir)


class PageOrganiser(object):
    def __init__(self, data: Dict[str, pd.DataFrame], grid_spec: tuple = None):
        if grid_spec is None:
            grid_spec = (1, 1)
        title_fontsize = "x-small"
        # self.fig = plt.figure(dpi=300, tight_layout=True)
        self.fig = plt.figure(dpi=300)
        self.fig.set_size_inches(8.27, 11.69, forward=True)
        plt.figtext(0.02, 0.99, "", fontsize="small")
        self.gs = plt.GridSpec(*grid_spec, wspace=0.3, hspace=0.3)
        self.data = {
            k: df.reset_index().sort_values("group")
            if hasattr(df, "reset_index")
            else df.sort_values("group")
            for k, df in data.items()
        }

    def place_plot(self, func, xloc=None, yloc=None, *args, **kwargs):
        if xloc is None:
            xloc = slice(0, self.gs.ncols)
        if yloc is None:
            yloc = slice(0, self.gs.nrows)

        if not isinstance(func, list):
            func = [func]

        for f in func:
            f(
                *args,
                ax=self.fig.add_subplot(self.gs[xloc, yloc]),
                **kwargs,
            )

    def plot(self):
        instructions: Iterable[Dict[str, Union[str, Iterable]]] = (
            # [
            {
                "data": "slice",
                "func": "barplot",
                "args": ("ntraps", "position"),
                "kwargs": {"hue": "group", "palette": "muted"},
                "loc": (0, 0),
            },
            # {
            #     "data": "slice",
            #     "func": "barplot",
            #     "args": ("ntraps", "position"),
            #     "kwargs": {"hue": "group", "palette": "muted"},
            #     "loc": (0, 0),
            # },
            # ],
            {
                "data": "delta_traps",
                "func": "barplot",
                "args": ("axis", "value"),
                "kwargs": {
                    "hue": "group",
                    # "hue": "group",
                },
                "loc": (0, 1),
            },
            {
                "data": "slices",
                "func": "violinplot",
                "args": ("group", "median"),
                "kwargs": {
                    "hue": "timepoint",
                },
                "loc": (1, 0),
            },
            {
                "data": "pertrap_metric",
                "func": "histplot",
                "args": (0, None),
                "kwargs": {
                    "hue": "group",
                    # "hue": "group",
                },
                "loc": (1, 1),
            },
        )

        self.plots = [
            self.place_plot(
                self.gen_sns_wrapper(how),
                *how["loc"],
            )
            for how in instructions
        ]

    def show(self):
        plt.show()

    def save(self, path: PosixPath):
        pp = PdfPages(path)
        pp.savefig(self.fig)
        pp.close()

    def gen_sns_wrapper(self, how):
        def sns_wrapper(ax):
            getattr(sns, how["func"])(
                data=self.data[how["data"]],
                x=how["args"][0],
                y=how["args"][1],
                **how["kwargs"],
                ax=ax,
            )

        return sns_wrapper


# for dir in dirs:
#     print(f"Compiling {dir}")
#     fullpath = Path("/home/alan/Documents/dev/skeletons/data/") / dir / dir
#     compiler = ExperimentCompiler(None, fullpath)
#     tmp = compiler.run()
#     po = PageOrganiser(tmp, grid_spec=(3, 2))
#     po.plot()
#     po.save(fullpath / f"{dir.rstrip('/')}_report.pdf")

#### Live editing

# plot.set_title("Trap identification robustness")
# plot.set_xlabel("Axis")
# plot.set_ylabel("Distance (pixels)")
# plt.show()


# fig.align_labels()  # same as fig.align_xlabels(); fig.align_ylabels()


import numpy as np


# plt1 = dummyplot()
# plt2 = dummyplot()
# # df = (
# #     pd.DataFrame({ax: signal[:, i] for i, ax in enumerate(("x", "y"))})
# #     .reset_index()
# #     .melt("index")
# # )
# # sns.lineplot(data=df, x="index", y="value", hue="variable")

# # inverted_errors = {
# #     k: {pos: v[k] for pos, v in errors.items()} for k in list(errors.values())[0].keys()
# # }
# # for_addition = {
# #     k: [v[pos] for pos in merged.index.get_level_values("position")]
# #     for k, v in inverted_errors.items()
# # }
# # for k, v in for_addition.items():
# #     merged[k] = v

# # fig, axes = plt.subplots(2, 1, sharex=True)
# # for i in range(2):
# #     axes[i].plot(signal[:, i])
# # plt.show()

# f, ax = plt.subplots(figsize=(6, 15))
# df = df.sort_values("group")
# sns.barplot(data=df, x="ntraps", y="position", hue="group", palette="pastel")
# sns.set_color_codes("muted")
# sns.barplot(data=df, x="count", y="position", hue="group", palette="muted")
# plt.show()

# fpath = "/home/alan/Documents/dev/skeletons/data/2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01/2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01/URA7_young001.h5"
# with h5py.File(fpath, "r") as f:
#     x, y = zip(*f["trap_info/trap_locations"][()])

# plt.scatter(x, y)
# plt.show()
# sns.heatmap()
